/*
 * test_nx.c
 */

static char *rcsid = "$Header: /nfs/globus1/src/master/nexus/test/test_nx.c,v 1.31 1998/04/10 16:14:08 toonen Exp $";

#undef ARGS_TEST
#undef CONDITION_TEST
#define LOCAL_COMM_TEST
#undef ATTACH_COMM_TEST
#undef GRAM_MYJOB_COMM_TEST
#undef TIME_NEXUS_POLL
#undef USE_THREADED_HANDLER
#undef GET_ARGUMENTS_AUTOMATICALLY
   
#include "nexus.h"
#include "stdio.h"

#ifdef ARGS_TEST
static void args_test(int argc, char **argv);
#endif /* ARGS_TEST */

#ifdef CONDITION_TEST
static void condition_test(void);
#endif /* CONDITION_TEST */

#ifdef LOCAL_COMM_TEST
#ifndef COMM_TEST
#define COMM_TEST
#endif
static void local_comm_test(void);
#endif /* LOCAL_COMM_TEST */

#ifdef TIME_NEXUS_POLL
static void time_nexus_poll(void);
#endif /* TIME_NEXUS_POLL */

#ifdef GRAM_MYJOB_COMM_TEST
#include "gram_myjob.h"
static void gram_myjob_comm_test(int *argc, char *** argv);
#ifndef COMM_TEST
#define COMM_TEST
#endif
#endif /* GRAM_MYJOB_COMM_TEST */

#ifdef ATTACH_COMM_TEST
#ifndef COMM_TEST
#define COMM_TEST
#endif
static nexus_bool_t		attach_server_arg = NEXUS_FALSE;
static nexus_bool_t		attach_client_arg = NEXUS_FALSE;
static char *			attach_client_url;
static nexus_bool_t		attach_client_leave_server = NEXUS_FALSE;
static nexus_bool_t		attach_client_die = NEXUS_FALSE;

static void attach_client(void);
static void attach_server(void);

#endif /* ATTACH_COMM_TEST */


#ifdef COMM_TEST

static void comm_test(nexus_startpoint_t *startpoint);

static void comm_test_remote_handler(nexus_endpoint_t *endpoint,
				nexus_buffer_t *buffer,
				nexus_bool_t called_from_non_threaded_handler);
#define COMM_TEST_REMOTE_HANDLER_ID 0
static void comm_test_finish_handler(nexus_endpoint_t *endpoint,
				nexus_buffer_t *buffer,
				nexus_bool_t called_from_non_threaded_handler);
#define COMM_TEST_FINISH_HANDLER_ID 1


#ifdef USE_THREADED_HANDLER
#define COMM_TEST_HANDLER_TYPE NEXUS_HANDLER_TYPE_THREADED
#else
#define COMM_TEST_HANDLER_TYPE NEXUS_HANDLER_TYPE_NON_THREADED
#endif

static nexus_handler_t comm_test_remote_handler_table[] =
{
  {COMM_TEST_HANDLER_TYPE,
   comm_test_remote_handler},
  {NEXUS_HANDLER_TYPE_NON_THREADED,
   comm_test_finish_handler},
  {NEXUS_HANDLER_TYPE_NON_THREADED,
   (nexus_handler_func_t) NULL},
};

#define COMM_TEST_REMOTE_HANDLER_TABLE_SIZE 2

static void comm_test_reply_handler(nexus_endpoint_t *endpoint,
				nexus_buffer_t *buffer,
				nexus_bool_t called_from_non_threaded_handler);
#define COMM_TEST_REPLY_HANDLER_ID 0

static nexus_handler_t comm_test_reply_handler_table[] =
{
  {NEXUS_HANDLER_TYPE_NON_THREADED,
   comm_test_reply_handler},
  {NEXUS_HANDLER_TYPE_NON_THREADED,
   (nexus_handler_func_t) NULL},
};

#define COMM_TEST_REPLY_HANDLER_TABLE_SIZE 1


static void comm_test_finish_init();
static void comm_test_finish_destroy();
static void comm_test_finish_wait();
static void comm_test_finish_send(nexus_startpoint_t *sp);

#endif /* COMM_TEST */


#ifndef MAX
#define MAX(V1,V2) (((V1) > (V2)) ? (V1) : (V2))
#define MIN(V1,V2) (((V1) < (V2)) ? (V1) : (V2))
#endif


/*
 * unknown_handler()
 */
void
unknown_handler(nexus_endpoint_t *endpoint,
		nexus_buffer_t *buffer,
		int handler_id)
{
    nexus_printf("unknown_handler(): handler_id=%d\n", handler_id);
    nexus_printf("unknown_handler(): exiting\n");
} /* unknown_handler() */


/*
 * fault_callback()
 */
int fault_callback(void *user_arg, int fault_code)
{
    nexus_printf("fault_callback(): fault_code=%d\n",
		 fault_code);
    return(0);
} /* fault_callback() */


/*****************************************************************
 *		MAIN
 *****************************************************************/

/*
 * main()
 */
int main(int argc, char **argv)
{
    int i;
    int rc;
#ifdef GET_ARGUMENTS_AUTOMATICALLY
    int my_argc = 0;
    char **my_argv;
#else
    int my_argc = argc;
    char **my_argv = argv;
#endif

    rc = nexus_init(&my_argc,
		    &my_argv,
		    "NEXUS_ARGS",
		    "nx",
		    NULL);
    
    if (rc != NEXUS_SUCCESS && rc != NEXUS_ERROR_ALREADY_INITIALIZED)
    {
	printf("nexus_init() failed with rc=%d\n", rc);
	exit(0);
    }

    nexus_printf("main(): After nexus_init()\n");

    i = 1;
    while (i < my_argc)
    {
	if (   (strcmp(my_argv[i], "-h") == 0)
	    || (strcmp(my_argv[i], "-help") == 0)
	    || (strcmp(my_argv[i], "--h") == 0)
	    || (strcmp(my_argv[i], "--help") == 0) )
	{
	    printf("Usage: %s <args> -nx <nexus-args>\n", my_argv[0]);
	    printf("Where <args> is:\n");
#ifdef ATTACH_COMM_TEST
	    printf("    -s                        : Attach test server\n");
	    printf("    -c <URL>                  : Attach test client\n");
	    printf("                                <URL> is server's URL\n");
	    printf("    -cl                       : Client leaves server running\n");
	    printf("    -cdie                     : Force client failure\n");
#endif /* ATTACH_COMM_TEST */
	    printf("Where <nexus-args> is:\n");
#ifdef ARGS_TEST
	    printf("    -test <value>             :    Argument for test_nx\n");
#endif /* ARGS_TEST */
	    nexus_arguments_print();
	    goto alldone;
	}
	
#ifdef ATTACH_COMM_TEST
	else if (strcmp(my_argv[i], "-s") == 0)
	{
	    attach_server_arg = NEXUS_TRUE;
	    i++;
	}
	else if ((strcmp(my_argv[i], "-c") == 0) && (i + 1 < my_argc))
	{
	    attach_client_arg = NEXUS_TRUE;
	    attach_client_url = my_argv[i + 1];
	    i += 2;
	}
	else if (strcmp(my_argv[i], "-cl") == 0)
	{
	    attach_client_leave_server = NEXUS_TRUE;
	    i++;
	}
	else if (strcmp(my_argv[i], "-cdie") == 0)
	{
	    attach_client_die = NEXUS_TRUE;
	    i++;
	}
#endif /* ATTACH_COMM_TEST */
	else
	{
	    /* Ignore unrecognized arguments */
	    i++;
	}
    }
    
    nexus_enable_fault_tolerance(fault_callback, NULL);
    
#ifdef ARGS_TEST
    args_test(my_argc, my_argv);
#endif /* ARGS_TEST */    
    
#ifdef CONDITION_TEST
    condition_test();
#endif /* CONDITION_TEST */

#ifdef LOCAL_COMM_TEST
    local_comm_test();
#endif /* LOCAL_COMM_TEST */

#ifdef GRAM_MYJOB_COMM_TEST
    gram_myjob_comm_test(&my_argc, &my_argv);
#endif /* GRAM_MYJOB_COMM_TEST */
    
#ifdef ATTACH_COMM_TEST
    if (attach_server_arg)
    {
	attach_server();
	goto alldone;
    }
    else if (attach_client_arg)
    {
	attach_client();
    }
    else
    {
	nexus_printf("Neither -s or -c was specified.  Skipping attach test\n");
    }
#endif /* ATTACH_COMM_TEST */
    
#ifdef TIME_NEXUS_POLL
    time_nexus_poll();
#endif /* TIME_NEXUS_POLL */

 alldone:
    
    nexus_printf("main(): calling nexus_shutdown()\n");
    nexus_shutdown();
    return(0);

} /* main() */


/*****************************************************************
 *		ARGS_TEST
 *****************************************************************/

#ifdef ARGS_TEST    
static void args_test(int argc, char **argv)
{
    int i;
    int *retained_argc;
    char ***retained_argv;
    int arg_num;

    nexus_arguments_get_retained(&retained_argc, &retained_argv);
    if ((arg_num = nexus_arguments_find(retained_argc,
					retained_argv,
					"test", 2)) >= 0)
    {
	nexus_printf("args_test(): Got -test argument, value=%s\n",
		     (*retained_argv)[arg_num+1]);
	nexus_arguments_remove(retained_argc,
			       retained_argv,
			       arg_num, 2);
    }
    
    nexus_printf("args_test(): argc=%d\n", argc);
    for (i = 0; i < argc; i++)
    {
	nexus_printf("args_test(): argv[%d]=%s\n", i, argv[i]);
    }
}
#endif /* ARGS_TEST */


/*****************************************************************
 *		CONDITION_TEST
 *****************************************************************/
#ifdef CONDITION_TEST

static nexus_mutex_t	condition_test_mutex;
static nexus_mutex_t    busy_mutex;
static nexus_cond_t	condition_test_cond;
/*
static int  cond_flag1 = 0;
static int  cond_flag2 = 0;
*/
void *condition_test_thread2(void *arg)
{
    int rc;

    nexus_printf("condition_test(): thread2: entering\n");

    nexus_mutex_lock(&condition_test_mutex);
    nexus_printf("condition_test(): thread2: waiting for signal \n");
    nexus_cond_wait(&condition_test_cond, &condition_test_mutex);
    nexus_printf("condition_test(): thread2: awoken\n");
    nexus_mutex_unlock(&condition_test_mutex);
    
    nexus_mutex_lock(&condition_test_mutex);
    nexus_printf("condition_test(): thread2: waiting for broadcast\n");
    nexus_cond_wait(&condition_test_cond, &condition_test_mutex);
    nexus_printf("condition_test(): thread2: awoken\n");
    nexus_mutex_unlock(&condition_test_mutex);

    nexus_mutex_lock(&condition_test_mutex);
    nexus_printf("condition_test(): thread2: waiting for sync\n");
    nexus_cond_wait(&condition_test_cond, &condition_test_mutex);
    nexus_mutex_unlock(&condition_test_mutex);
    nexus_printf("condition_test(): thread2: trying to acquire busy lock\n");
    rc = nexus_mutex_trylock(&busy_mutex);
    if (rc != EBUSY)
    {
	nexus_fatal("condition_test(): lock returned %d instead of %d (EBUSY)\n", rc, EBUSY);
    }
    nexus_cond_signal(&condition_test_cond);
    nexus_printf("condition_test(): thread2: busy lock not acquired\n");

    nexus_printf("condition_test(): thread2: exiting\n");

    return (NULL);
} /* condition_test_thread2() */

/*
 * condition_test()
 */
static void condition_test(void)
{
    nexus_thread_t thread;
    int rc;
    
    nexus_printf("condition_test(): starting\n");
    
    nexus_mutex_init(&condition_test_mutex, (nexus_mutexattr_t *) NULL);
    nexus_mutex_init(&busy_mutex, (nexus_mutexattr_t *) NULL);
    nexus_cond_init(&condition_test_cond, (nexus_condattr_t *) NULL);
    
    nexus_thread_create(&thread,
			(nexus_thread_attr_t *) NULL,
			condition_test_thread2,
			(void *) NULL);

    sleep(3); nexus_thread_yield();
    
    nexus_mutex_lock(&condition_test_mutex);
    nexus_printf("condition_test(): thread1: signaling\n");
    nexus_mutex_unlock(&condition_test_mutex);
    nexus_cond_signal(&condition_test_cond);
    nexus_mutex_lock(&condition_test_mutex);
    nexus_printf("condition_test(): thread1: done signaling\n");
    nexus_mutex_unlock(&condition_test_mutex);

    sleep(3); nexus_thread_yield();
    
    nexus_mutex_lock(&condition_test_mutex);
    nexus_printf("condition_test(): thread1: broadcasting\n");
    nexus_mutex_unlock(&condition_test_mutex);
    nexus_cond_broadcast(&condition_test_cond);
    nexus_mutex_lock(&condition_test_mutex);
    nexus_printf("condition_test(): thread1: done broadcasting\n");
    nexus_mutex_unlock(&condition_test_mutex);

    sleep(3); nexus_thread_yield();

    nexus_printf("condition_test(): thread1: trying to acquire busy lock\n");
    rc = nexus_mutex_trylock(&busy_mutex);
    if (rc != 0)
    {
	nexus_fatal("condition_test(): trylock returned %d instead of 0 (success)\n", rc);
    }
    nexus_printf("condition_test(): thread1: busy lock acquired\n");
    nexus_cond_signal(&condition_test_cond);
    nexus_mutex_lock(&condition_test_mutex);
    nexus_printf("condition_test(): thread1: waiting for thread2\n");
    nexus_cond_wait(&condition_test_cond, &condition_test_mutex);
    nexus_mutex_unlock(&condition_test_mutex);
    nexus_printf("condition_test(): thread1: unlocking busy lock\n");
    nexus_mutex_unlock(&busy_mutex);

    sleep(3); nexus_thread_yield();

    nexus_mutex_destroy(&condition_test_mutex);
    nexus_mutex_destroy(&busy_mutex);
    nexus_cond_destroy(&condition_test_cond);
    
    nexus_printf("condition_test(): complete\n");
} /* condition_test() */

#endif /* CONDITION_TEST */


/*****************************************************************
 *		COMM_TEST
 *****************************************************************/
#ifdef COMM_TEST

typedef struct _comm_test_reply_done_t
{
    nexus_mutex_t	mutex;
    nexus_cond_t	cond;
    nexus_bool_t		done;
} comm_test_reply_done_t;
static comm_test_reply_done_t comm_test_reply_done;

static int			comm_test_int_data = 42;


/*
 * comm_test()
 *
 * Invoke comm_test_remote_handler() in the context pointed to
 * by 'remote_sp'.  The remote handler will bounce a message
 * back to the original context using comm_test_reply_handler().
 * The reply handler will signal the original thread via a monitor.
 */
static void comm_test(nexus_startpoint_t *remote_sp)
{
    nexus_endpointattr_t reply_epattr;
    nexus_endpoint_t reply_ep;
    nexus_startpoint_t reply_sp;
    nexus_buffer_t buffer;
    int buf_size;
    
    nexus_printf("comm_test(): starting\n");

    /*
     * Initialize the monitor that will be used for signaling
     * completion of the reply handler.
     */
    nexus_mutex_init(&(comm_test_reply_done.mutex),
		     (nexus_mutexattr_t *) NULL);
    nexus_cond_init(&(comm_test_reply_done.cond), (nexus_condattr_t *) NULL);
    nexus_mutex_lock(&(comm_test_reply_done.mutex));
    comm_test_reply_done.done = NEXUS_FALSE;
    nexus_mutex_unlock(&(comm_test_reply_done.mutex));

    /*
     * Set up a reply endpoint/startpoint, and point
     * it at the reply done monitor
     */
    nexus_endpointattr_init(&reply_epattr);
    nexus_endpointattr_set_handler_table(&reply_epattr,
					 comm_test_reply_handler_table,
					 COMM_TEST_REPLY_HANDLER_TABLE_SIZE);
    nexus_endpointattr_set_unknown_handler(&reply_epattr,
					   unknown_handler,
					   NEXUS_HANDLER_TYPE_NON_THREADED);
    nexus_endpoint_init(&reply_ep, &reply_epattr);
    nexus_endpoint_set_user_pointer(&reply_ep,
				    (void *)&comm_test_reply_done);
    nexus_startpoint_bind(&reply_sp, &reply_ep);

    /*
     * Test nexus_startpoint_*() equality functions
     */
    if (nexus_startpoint_to_current_context(&reply_sp))
    {
	nexus_endpoint_t *tmp_ep;
	nexus_printf("comm_test(): nexus_startpoint_to_current_context() works\n");
	nexus_startpoint_get_endpoint(&reply_sp, &tmp_ep);
	if (tmp_ep && (tmp_ep == &reply_ep))
	{
	    nexus_printf("comm_test(): nexus_startpoint_get_endpoint() works\n");
	}
	else
	{
	    nexus_printf("comm_test(): ERROR: nexus_startpoint_get_endpoint() failed\n");
	}
    }
    else
    {
	nexus_printf("comm_test(): ERROR: nexus_startpoint_to_current_context() failed\n");
    }
    
    /*
     * Do the RSR
     */
    buf_size = nexus_sizeof_startpoint(&reply_sp, 1);
    buf_size += nexus_sizeof_int(1);
    nexus_buffer_init(&buffer, buf_size, 0);
    nexus_put_startpoint_transfer(&buffer, (&reply_sp), 1);
    nexus_put_int(&buffer, &comm_test_int_data, 1);

    nexus_printf("comm_test(): sending rsr\n");
    if (nexus_send_rsr(&buffer,
		       remote_sp,
		       COMM_TEST_REMOTE_HANDLER_ID,
		       NEXUS_TRUE,
		       NEXUS_FALSE) != NEXUS_SUCCESS)
    {
	nexus_printf("comm_test(): WARNING: send failed\n");
    }

#ifdef ATTACH_COMM_TEST
    /*
     * Die (exit) unexpectedly to test fault tolerance
     */
    if (attach_client_die)
    {
	nexus_printf("comm_test(): client simulating death\n");
	exit(0);
    }
#endif /* ATTACH_COMM_TEST */

    /*
     * Suspend main thread until handler completes
     */
    nexus_printf("comm_test(): entering monitor\n");
    nexus_mutex_lock(&(comm_test_reply_done.mutex));
    nexus_printf("comm_test(): waiting on monitor\n");
    while (!comm_test_reply_done.done)
    {
	nexus_cond_wait(&(comm_test_reply_done.cond),
			&(comm_test_reply_done.mutex));
    }
    nexus_printf("comm_test(): awoke from monitor and exiting\n");
    nexus_mutex_unlock(&(comm_test_reply_done.mutex));

    nexus_mutex_destroy(&(comm_test_reply_done.mutex));
    nexus_cond_destroy(&(comm_test_reply_done.cond));
    nexus_startpoint_destroy(&reply_sp);
    nexus_endpoint_destroy(&reply_ep);
    nexus_endpointattr_destroy(&reply_epattr);
    
    nexus_printf("comm_test(): complete\n");
    
} /* comm_test() */


/*
 * comm_test_remote_handler()
 */
static void comm_test_remote_handler(nexus_endpoint_t *endpoint,
				 nexus_buffer_t *buffer,
				 nexus_bool_t called_from_non_threaded_handler)
{
    int int_data;
    nexus_startpoint_t reply_sp;
    nexus_buffer_t reply_buffer;
    
    nexus_printf("comm_test_remote_handler(): entering\n");

    /* Get the reply global pointer */
    nexus_get_startpoint(buffer, &reply_sp, 1);
    
    /* Test nexus_startpoint_*() equality functions */
    nexus_printf("comm_test(): nexus_startpoint_to_current_context(&reply_sp)==%s (should be TRUE for local_comm_test(), and FALSE for attach_comm_test(), and depentent upon the context for gram_myjob_comm_test())\n", (nexus_startpoint_to_current_context(&reply_sp) ? "TRUE" : "FALSE"));
    
    /* Get the data and check to see if it made it ok */
    nexus_get_int(buffer, &int_data, 1);
    if (int_data == comm_test_int_data)
    {
	nexus_printf("comm_test_remote_handler(): got int_data ok\n");
    }
    else
    {
	nexus_printf("comm_test_remote_handler(): ERROR: bad int_data %d, should be %d\n",
		     int_data, comm_test_int_data);
    }

    if (!called_from_non_threaded_handler)
    {
        nexus_printf("comm_test_remote_handler(): freeing saved buffer\n");
	nexus_buffer_destroy(buffer);
    }
    
    /* Reply to the original context */
    nexus_printf("comm_test_remote_handler(): sending reply rsr\n");
    nexus_buffer_init(&reply_buffer, 0, 0);
    if (nexus_send_rsr(&reply_buffer,
		       &reply_sp,
		       COMM_TEST_REPLY_HANDLER_ID,
		       NEXUS_TRUE,
		       called_from_non_threaded_handler) != NEXUS_SUCCESS)
    {
	nexus_printf("comm_test_remote_handler(): WARNING: send failed\n");
	nexus_printf("comm_test_remote_handler(): nexus_startpoint_test()=%d\n",
		     nexus_startpoint_test(&reply_sp));
    }

} /* comm_test_remote_handler() */


/*
 * comm_test_reply_handler()
 */
static void comm_test_reply_handler(nexus_endpoint_t *endpoint,
				 nexus_buffer_t *buffer,
				 nexus_bool_t called_from_non_threaded_handler)
{
    struct _comm_test_reply_done_t *reply_done;
    void *address;
    
    nexus_printf("comm_test_reply_handler(): entering\n");

    address = nexus_endpoint_get_user_pointer(endpoint);
    /* The address should be to the comm_test_reply_done structure */
    if (address == (void *) (&comm_test_reply_done))
    {
	reply_done = (comm_test_reply_done_t *) address;
	nexus_printf("comm_test_reply_handler(): address ok\n");
    }
    else
    {
	reply_done = &comm_test_reply_done;
	nexus_printf("comm_test_reply_handler(): ERROR: bad address %x, should be %x\n",
		     address,
		     (&comm_test_reply_done));
    }
    
    /* Awaken the main thread */
    nexus_printf("comm_test_reply_handler(): entering monitor\n");
    nexus_mutex_lock(&(reply_done->mutex));
    reply_done->done = NEXUS_TRUE;
    nexus_printf("comm_test_reply_handler(): signaling monitor\n");
    nexus_cond_signal(&(reply_done->cond));
    nexus_mutex_unlock(&(reply_done->mutex));
   
    nexus_printf("comm_test_reply_handler(): exiting\n");

} /* comm_test_reply_handler() */



typedef struct
{
    nexus_mutex_t	mutex;
    nexus_cond_t	cond;
    nexus_bool_t	done;
} comm_test_finish_monitor_t;

static comm_test_finish_monitor_t comm_test_finish_monitor;

/*
 * comm_test_finish_init()
 */
static void comm_test_finish_init()
{
    nexus_mutex_init(&(comm_test_finish_monitor.mutex),
		     (nexus_mutexattr_t *) NULL);
    nexus_cond_init(&(comm_test_finish_monitor.cond),
		    (nexus_condattr_t *) NULL);
    comm_test_finish_monitor.done = NEXUS_FALSE;
} /* comm_test_finish_init() */


/*
 * comm_test_finish_destroy()
 */
static void comm_test_finish_destroy()
{
    nexus_mutex_destroy(&(comm_test_finish_monitor.mutex));
    nexus_cond_destroy(&(comm_test_finish_monitor.cond));
} /* comm_test_finish_destroy() */


/*
 * comm_test_finish_wait()
 */
static void comm_test_finish_wait()
{
    nexus_printf("comm_test_finish_wait(): waiting on monitor\n");
    nexus_mutex_lock(&(comm_test_finish_monitor.mutex));
    while (!(comm_test_finish_monitor.done))
    {
	nexus_cond_wait(&(comm_test_finish_monitor.cond),
			&(comm_test_finish_monitor.mutex));
    }
    nexus_mutex_unlock(&(comm_test_finish_monitor.mutex));
    nexus_printf("gram_myjob_finish_wait(): awoke from monitor\n");
} /* comm_test_finish_wait() */


/*
 * comm_test_finish_send()
 */
static void comm_test_finish_send(nexus_startpoint_t *sp)
{
    nexus_buffer_t buffer;
    
    nexus_printf("comm_test_finish_send(): sending finish message\n");
    
    nexus_buffer_init(&buffer, 0, 0);
    if (nexus_send_rsr(&buffer, sp,
		       COMM_TEST_FINISH_HANDLER_ID,
		       NEXUS_TRUE, NEXUS_FALSE) != 0)
    {
	nexus_printf("comm_test_finish_send(): WARNING: send failed\n");
    }
} /* comm_test_finish_send() */


/*
 * comm_test_finish_handler()
 */
static void comm_test_finish_handler(nexus_endpoint_t *endpoint,
				 nexus_buffer_t *buffer,
				 nexus_bool_t called_from_non_threaded_handler)
{
    nexus_printf("comm_test_finish_handler(): signaling\n");
    nexus_mutex_lock(&(comm_test_finish_monitor.mutex));
    comm_test_finish_monitor.done = NEXUS_TRUE;
    nexus_cond_signal(&(comm_test_finish_monitor.cond));
    nexus_mutex_unlock(&(comm_test_finish_monitor.mutex));
    nexus_printf("comm_test_finish_handler(): done signaling\n");
} /* comm_test_finish_handler() */

#endif /* COMM_TEST */



/*****************************************************************
 *		LOCAL_COMM_TEST
 *****************************************************************/
#ifdef LOCAL_COMM_TEST

/*
 * local_comm_test()
 */
static void local_comm_test(void)
{
    nexus_endpointattr_t epattr;
    nexus_endpoint_t ep;
    nexus_startpoint_t sp;
    
    nexus_printf("local_comm_test(): starting\n");

    nexus_endpointattr_init(&epattr);
    nexus_endpointattr_set_handler_table(&epattr,
					 comm_test_remote_handler_table,
					 COMM_TEST_REMOTE_HANDLER_TABLE_SIZE);
    nexus_endpointattr_set_unknown_handler(&epattr,
					   unknown_handler,
					   NEXUS_HANDLER_TYPE_NON_THREADED);
    nexus_endpoint_init(&ep, &epattr);
    nexus_startpoint_bind(&sp, &ep);
    
    comm_test(&sp);
    
    nexus_startpoint_destroy(&sp);
    nexus_endpoint_destroy(&ep);
    nexus_endpointattr_destroy(&epattr);
    
    nexus_printf("local_comm_test(): complete\n");
} /* local_comm_test() */

#endif /* LOCAL_COMM_TEST */



/*****************************************************************
 *		GRAM_MYJOB_COMM_TEST
 *****************************************************************/
#ifdef GRAM_MYJOB_COMM_TEST

/*
 * gram_myjob_comm_test()
 */
static void gram_myjob_comm_test(int *argc, char ***argv)
{
    nexus_endpointattr_t epattr;
    nexus_endpoint_t ep;
    nexus_startpoint_t sp;
    int rc;
    int job_size;
    int rank;
    nexus_byte_t array[GRAM_MYJOB_MAX_BUFFER_LENGTH];
    nexus_byte_t *a;
    int length;
    
    nexus_printf("gram_myjob_comm_test(): starting\n");

    /* Initialize gram_myjob */
    if ((rc = gram_myjob_init(argc, argv)) != GRAM_SUCCESS)
    {
	nexus_printf("gram_myjob_comm_test(): ERROR: gram_myjob_init() returned %d\n", rc);
	return;
    }

    /* Get my rank and the job size */
    if ((rc = gram_myjob_size(&job_size)) != GRAM_SUCCESS)
    {
	nexus_printf("gram_myjob_comm_test(): ERROR: gram_myjob_size() returned %d\n", rc);
	return;
    }
    if ((rc = gram_myjob_rank(&rank)) != GRAM_SUCCESS)
    {
	nexus_printf("gram_myjob_comm_test(): ERROR: gram_myjob_rank() returned %d\n", rc);
	return;
    }
    nexus_printf("gram_myjob_comm_test(): size=%d, rank=%d\n", job_size, rank);

    /* Create an endpoint */
    nexus_endpointattr_init(&epattr);
    nexus_endpointattr_set_handler_table(&epattr,
					 comm_test_remote_handler_table,
					 COMM_TEST_REMOTE_HANDLER_TABLE_SIZE);
    nexus_endpointattr_set_unknown_handler(&epattr,
					   unknown_handler,
					   NEXUS_HANDLER_TYPE_NON_THREADED);
    nexus_endpoint_init(&ep, &epattr);

    if (rank == 0)
    {
	nexus_startpoint_t *sp_array;
	int format;
	int source;
	int i;

	sp_array = (nexus_startpoint_t *)
	    nexus_malloc(sizeof(nexus_startpoint_t) * job_size);

	nexus_startpoint_bind(&(sp_array[0]), &ep);
	
	for (i = 1; i < job_size; i++)
	{
	    nexus_printf("gram_myjob_comm_test(): Calling gram_myjob_receive()\n");
	    if ((rc = gram_myjob_receive(array, &length)) != GRAM_SUCCESS)
	    {
		nexus_printf("gram_myjob_comm_test(): ERROR: gram_myjob_send() returned %d\n", rc);
		return;
	    }
	    nexus_printf("gram_myjob_comm_test(): gram_myjob_receive() returned length=%d\n", length);
	    
	    a = array;
	    format = (int) (*a++);
	    nexus_user_get_int(&a, &source, 1, format);
	    if (source < 1 || source >= job_size)
	    {
		nexus_printf("gram_myjob_comm_test(): ERROR: got invalid source = %d\n", source);
		nexus_shutdown();
		exit(1);
	    }
	    nexus_user_get_startpoint(&a, &(sp_array[source]), 1, format);
	    nexus_printf("gram_myjob_comm_test(): Got startpoint()\n");
	}

	/* Shutdown gram_myjob */
	nexus_printf("gram_myjob_comm_test(): Calling gram_myjob_done()\n");
	if ((rc = gram_myjob_done()) != GRAM_SUCCESS)
	{
	    nexus_printf("gram_myjob_comm_test(): ERROR: gram_myjob_done() returned %d\n", rc);
	    return;
	}

	for (i = 0; i < job_size; i++)
	{
	    nexus_printf("gram_myjob_comm_test(): running comm_test() to job process %d\n", i);
	    comm_test(&(sp_array[i]));
	}

	
	nexus_startpoint_destroy(&(sp_array[0]));
	for (i = 1; i < job_size; i++)
	{
	    nexus_printf("gram_myjob_comm_test(): shutting down job process %d\n", i);
	    comm_test_finish_send(&(sp_array[i]));
	    nexus_startpoint_destroy(&(sp_array[i]));
	}

    }
    else
    {
	comm_test_finish_init();

	/* Send a startpoint to rank 0 */
	nexus_startpoint_bind(&sp, &ep);
	a = array;
	*a++ = (nexus_byte_t) nexus_user_format();
	nexus_user_put_int(&a, &rank, 1);
	nexus_user_put_startpoint_transfer(&a, &sp, 1);
	length = a - array;
	
	nexus_printf("gram_myjob_comm_test(): gram_myjob_send() sending length=%d\n", length);
	if ((rc = gram_myjob_send(0, array, length)) != GRAM_SUCCESS)
	{
	    nexus_printf("gram_myjob_comm_test(): ERROR: gram_myjob_send() returned %d\n", rc);
	    return;
	}

	/* Shutdown gram_myjob */
	nexus_printf("gram_myjob_comm_test(): Calling gram_myjob_done()\n");
	if ((rc = gram_myjob_done()) != GRAM_SUCCESS)
	{
	    nexus_printf("gram_myjob_comm_test(): ERROR: gram_myjob_done() returned %d\n", rc);
	    return;
	}

	/* Wait for the communication test to complete */
	comm_test_finish_wait();
	comm_test_finish_destroy();
    }

    nexus_endpoint_destroy(&ep);
    nexus_endpointattr_destroy(&epattr);
    
    nexus_printf("gram_myjob_comm_test(): complete\n");
} /* gram_myjob_comm_test() */

#endif /* GRAM_MYJOB_COMM_TEST */



/*****************************************************************
 *		ATTACH_COMM_TEST
 *****************************************************************/
#ifdef ATTACH_COMM_TEST

/*
 * attach_requested()
 */
int attach_requested(void *user_arg, char *url, nexus_startpoint_t *sp)
{
    nexus_endpoint_t *ep = (nexus_endpoint_t *) user_arg;
    char *host;
    unsigned short port;
    char **specs;
    int i;
    
    nexus_printf("attach_requested(): entering\n");
    nexus_printf("attach_requested(): url=%s\n", url);

    nexus_split_url(url, &host, &port, &specs);
    nexus_printf("attach_requested(): host=%s\n", host);
    nexus_printf("attach_requested(): port=%hu\n", port);
    for (i = 0; specs[i]; i++)
    {
	nexus_printf("attach_requested(): spec %d: <%s>\n", i, specs[i]);
    }
    nexus_split_url_free(&host, &specs);

    nexus_startpoint_bind(sp, ep);
    
    nexus_printf("attach_requested(): exiting\n");
    return(0);
} /* attach_requested() */


/*
 * attach_server()
 */
static void attach_server(void)
{
    nexus_endpointattr_t epattr;
    nexus_endpoint_t ep;
    nexus_startpoint_t sp;
    unsigned short port;
    char *host, *host2;
    int rc;
    
    nexus_printf("attach_server(): entering\n");

    /*
     * Create an endpoint that will be used for attachments
     */
    nexus_endpointattr_init(&epattr);
    nexus_endpointattr_set_handler_table(&epattr,
					 comm_test_remote_handler_table,
					 COMM_TEST_REMOTE_HANDLER_TABLE_SIZE);
    nexus_endpointattr_set_unknown_handler(&epattr,
					   unknown_handler,
					   NEXUS_HANDLER_TYPE_NON_THREADED);
    nexus_endpoint_init(&ep, &epattr);

    comm_test_finish_init();
    
    nexus_printf("attach_server(): allowing attachment\n");
    port = 0;
    if ((rc = nexus_allow_attach(&port, &host,
				 &attach_requested,
				 (void *) &ep)) != 0)
    {
	nexus_printf("ERROR: nexus_allow_attach() returned non-0: rc=%d\n",
		     rc);
    }
    if ((rc = nexus_allow_attach(&port, &host2, &attach_requested, NULL)) != 1)
    {
	nexus_printf("ERROR: second nexus_allow_attach() did not return 1: rc=%d\n\n", rc);
    }
    
    nexus_printf("attach_server(): ATTACH TO: x-nexus://%s:%hu\n",
		 host, port);

    comm_test_finish_wait();
    
    nexus_disallow_attach(port);

    comm_test_finish_destroy();
    nexus_endpoint_destroy(&ep);
    nexus_endpointattr_destroy(&epattr);

    nexus_printf("attach_server(): exiting\n");
    
} /* attach_server() */


/*
 * attach_client()
 */
static void attach_client(void)
{
    char url[1024];
    nexus_startpoint_t sp;
    nexus_buffer_t buffer;
    int rc;

    nexus_printf("attach_client(): entering\n");

    nexus_enable_fault_tolerance(fault_callback, NULL);

    sprintf(url, "%s/spec1//spec\\/3/spec4", attach_client_url);

    nexus_printf("attach_client(): attaching to url <%s>\n", url);
    rc = nexus_attach(url, &sp);
    if (rc == 0)
    {
	nexus_printf("attach_client(): running comm_test() to server\n");
	comm_test(&sp);
	nexus_printf("attach_client(): done running comm_test() to server\n");

	if (!attach_client_leave_server)
	{
	    /* Send rsr to shut the server down */
	    comm_test_finish_send(&sp);
	}

	nexus_startpoint_destroy(&sp);
    }
    else
    {
	nexus_printf("attach_client(): ERROR: nexus_attach() failed with rc=%d\n", rc);
    }
    
    nexus_printf("attach_client(): exiting\n");
    
} /* attach_client() */

#endif /* ATTACH_COMM_TEST */


/*****************************************************************
 *		TIME_NEXUS_POLL
 *****************************************************************/
#ifdef TIME_NEXUS_POLL

#define TIME_NEXUS_POLL_ITERS 10000

/*
 * time_nexus_poll()
 */
static void time_nexus_poll(void)
{
    double start, stop, elapse;
    int i;

    nexus_printf("time_nexus_poll(): starting\n");

    start = nexus_wallclock();
    for (i = 0; i < TIME_NEXUS_POLL_ITERS; i++)
    {
	nexus_poll();
    }
    stop = nexus_wallclock();
    elapse = stop - start;
    nexus_printf("time_nexus_poll(): iters=%d, elapse=%f, elapse/iters=%f\n",
		 TIME_NEXUS_POLL_ITERS,
		 elapse, 
		 (elapse/TIME_NEXUS_POLL_ITERS) );
    
    nexus_printf("time_nexus_poll(): complete\n");
} /* time_nexus_poll() */

#endif /* TIME_NEXUS_POLL */
