/******************************************************************************
CIF_Factory.C

Description:

  CIF Object Factory

CVS Information:

  $Source$
  $Date$
  $Revision$
  $State$
  $Author$
******************************************************************************/

/******************************************************************************
			     Include header files
******************************************************************************/
#include "CIF.H"
#include "CIF_Factory.H"
#include "CIF_Connection_Nexus_ROM.H"
#include "CIF_Connection_Nexus_TCP.H"
#include "CIF_Connection_Nexus_UDP.H"
#include "CIF_Connection_Nexus_UUM.H"
#include "CIF_Listener_Nexus_TCP.H"
#include "CIF_Listener_Nexus_UDP.H"
#include "globus_common.h"


/******************************************************************************
			Define methods for CIF_Factory
******************************************************************************/
CIF_Connection *
CIF_Factory::get_Connection(
    ARG_IN char *			url,
    ARG_IN int				timeout)
{
    CIF_Connection *			conn;
    globus_url_t			url_info;

    if (globus_url_parse(url, &url_info) != GLOBUS_SUCCESS)
    {
	CIF::status = CIF::STATUS_URL_FORMAT_BAD;
	return CIF_NULL;
    }

    if (strcmp(url_info.scheme,
	       CIF_PROTOCOL_ROM_STR) == 0)
    {
	conn = (CIF_Connection *) new CIF_Connection_Nexus_ROM();
    }
    else if (strcmp(url_info.scheme,
	       CIF_PROTOCOL_TCP_STR) == 0)
    {
	conn = (CIF_Connection *) new CIF_Connection_Nexus_TCP();
    }
    else if (strcmp(url_info.scheme,
	       CIF_PROTOCOL_UDP_STR) == 0)
    {
	conn = (CIF_Connection *) new CIF_Connection_Nexus_UDP();
    }
    else if (strcmp(url_info.scheme,
	       CIF_PROTOCOL_UUM_STR) == 0)
    {
	conn = (CIF_Connection *) new CIF_Connection_Nexus_UUM();
    }
    else
    {
	CIF::status = CIF::STATUS_URL_SCHEME_UNSUPPORTED;
	return CIF_NULL;
    }
    
    if (Nexus::status != Nexus::STATUS_SUCCESS || !conn->open(url, timeout))
    {
	CIF::Status			status;

	status = CIF::status;
	delete conn;
	CIF::status = status;
	conn = CIF_NULL;
    }

    return conn;
}
    

CIF_Listener *
CIF_Factory::get_Listener(
    ARG_IN char *			url)
{
    CIF_Listener *			listener;
    int					port;
    
    if (strncmp(url, CIF_PROTOCOL_TCP_STR,
		strlen(CIF_PROTOCOL_TCP_STR)) == 0)
    {
	url += strlen(CIF_PROTOCOL_TCP_STR);
	listener = (CIF_Listener *) new CIF_Listener_Nexus_TCP();
    }
    else if (strncmp(url, CIF_PROTOCOL_UDP_STR,
		     strlen(CIF_PROTOCOL_UDP_STR)) == 0)
    {
	url += strlen(CIF_PROTOCOL_UDP_STR);
	listener = (CIF_Listener *) new CIF_Listener_Nexus_UDP();
    }
    else
    {
	CIF::status = CIF::STATUS_URL_SCHEME_UNSUPPORTED;
	return CIF_NULL;
    }

    if (url[0] != ':')
    {
	delete listener;
	CIF::status = CIF::STATUS_URL_FORMAT_BAD;
	return CIF_NULL;
    }

    port = -1;
    if (url[1] != 0)
    {
	char				tmp_char;
	
	if (sscanf(url, "://%d%c", &port, &tmp_char) != 1 || port < 0)
	{
	    delete listener;
	    CIF::status = CIF::STATUS_URL_FORMAT_BAD;
	    return CIF_NULL;
	}
    }
    
    if (Nexus::status != Nexus::STATUS_SUCCESS ||
	!((port < 0 && listener->open()) ||
	  (port >= 0 && listener->open(port))))
    {
	CIF::Status			status;

	status = CIF::status;
	delete listener;
	CIF::status = status;
	listener = CIF_NULL;
    }

    return listener;
}

