/******************************************************************************
CIF_Listener_Nexus.H

Description:

  Nexus Listener Class

CVS Information:

  $Source$
  $Date$
  $Revision$
  $State$
  $Author$
******************************************************************************/

#if !defined(CIF_INCLUDE_CIF_LISTENER_NEXUS)
#define CIF_INCLUDE_CIF_LISTENER_NEXUS 1

/******************************************************************************
			     Include header files
******************************************************************************/
#include "CIF_Listener.H"
#include "CIF_Connection_Nexus.H"
#include "Nexus_internal.H"

 
/******************************************************************************
			Define CIF_Listener_Nexus class
******************************************************************************/
class
CIF_Listener_Nexus
: public CIF_Listener
{
    /**************************************************************************
				    API methods
    **************************************************************************/

    public:
    CIF_Listener_Nexus();

    public: virtual
    CIF_Bool
    open(
	ARG_IN unsigned short		port = 0);

    public: virtual
    CIF_Connection *
    get_Connection(
	ARG_IN int			timeout = -1);

    public: virtual
    char *
    get_URL();

    public: virtual
    CIF_Bool
    close();

    public: virtual
    ~CIF_Listener_Nexus();
    
    /**************************************************************************
			      Derived API methods
    **************************************************************************/
    protected: virtual
    CIF_Connection_Nexus *
    allocate_Connection() = 0;

    protected: virtual
    char *
    get_URL_Scheme() = 0;

    
    /**************************************************************************
		       Internal methods and data structures
    **************************************************************************/
    enum Object_State
    {
	OBJECT_STATE_INACTIVE,
	OBJECT_STATE_ACTIVE,
	OBJECT_STATE_DELETED
    };
    
    private:
    CIF_Bool
    l_close();

    private:
    CIF::Status
    verify_Object_State(
	Object_State			desired_state);
    
    private:
    /* thread safety data structures */
    nexus_mutex_t			mutex;

    /* endpoint for receiving connection establishment information */
    Nexus_Endpoint			endpoint;

    /* listener for establishing the control connection */
    Nexus_Listener			listener;

    /* URL used to contact this listener */
    char *				url;
    
    /* state of the connection */
    volatile Object_State		object_state;
};

#endif /* CIF_INCLUDE_CIF_LISTENER_NEXUS */
