/******************************************************************************
CIF_Message.H

Description:

  Message buffer class

  Note: The message data buffer is completely exposed to the user.  Any
  modifications made to the data pointed to by get_Data() will permanently
  affect the data stored in the buffer.

CVS Information:

  $Source$
  $Date$
  $Revision$
  $State$
  $Author$
******************************************************************************/

#if !defined(CIF_INCLUDE_CIF_MESSAGE)
#define CIF_INCLUDE_CIF_MESSAGE 1

/******************************************************************************
			     Include header files
******************************************************************************/
#include "CIF.H"


/******************************************************************************
			       Type definitions
******************************************************************************/

/*
 * Buffer meta-data.  This will eventually contiain membershop information.
 */
struct CIF_Message_Info
{
    CIF_Size				size;
    int					tag;
};


/******************************************************************************
			       Class definition
******************************************************************************/
class
CIF_Message
{
    /**************************************************************************
				    API methods
    **************************************************************************/
    public:
    CIF_Message();

    public:
    CIF_Bool
    activate(
	const CIF_Message_Info &	info_arg);

    public:
    CIF_Byte *
    get_Data();

    public:
    CIF_Message_Info
    get_Info();

    public:
    CIF_Bool
    deactivate();
    
    public:
    ~CIF_Message();


    /**************************************************************************
			       Private data members
    **************************************************************************/
    private:

    /* Message data buffer */
    CIF_Byte *				data;

    /* Message meta-data */
    CIF_Message_Info			info;

    /* object state */
    enum Object_State
    {
	OBJECT_STATE_INACTIVE,
	OBJECT_STATE_ACTIVE,
	OBJECT_STATE_DELETED
    }					object_state;
};

#endif
