/******************************************************************************
Nexus_Monitor.C

Description:

  A simple monitor class built using nexus mutexes and condition variables

CVS Information:

  $Source$
  $Date$
  $Revision$
  $State$
  $Author$
******************************************************************************/

/******************************************************************************
			     Include header files
******************************************************************************/
#include "Nexus_Monitor.H"


/******************************************************************************
			  Define Nexus_Monitor class
******************************************************************************/

Nexus_Monitor::Nexus_Monitor(
    ARG_INMOD nexus_mutex_t *	mutex_arg,
    ARG_INMOD nexus_cond_t *	cond_arg)
: mutex(mutex_arg),
  cond(cond_arg)
{
    NexusAssert(mutex != GLOBUS_NULL);
    
    nexus_mutex_lock(mutex);
}

Nexus_Monitor::~Nexus_Monitor()
{
    NexusAssert(mutex != GLOBUS_NULL);

    nexus_mutex_unlock(mutex);
}

void
Nexus_Monitor::wait(
    ARG_IN const long			timeout)
{
    NexusAssert(mutex != GLOBUS_NULL);
    NexusAssert(cond != GLOBUS_NULL);

    if (timeout < 0)
    {
	nexus_cond_wait(cond, mutex);
    }
    else
    {
	/*
	 * TODO: This is really going to chew on the CPU, but it's the best
	 * we can do until we have timed events in the core of Nexus.  This
	 * code should be updated once timed events have been added.
	 */
	nexus_poll();
    }
}


void
Nexus_Monitor::signal()
{
    NexusAssert(cond != GLOBUS_NULL);

    nexus_cond_signal(&cond);
}

void
Nexus_Monitor::broadcast()
{
    NexusAssert(cond != GLOBUS_NULL);

    nexus_cond_broadcast(&cond);
}
